import typing
from unittest import mock

import httpx
import pytest
from pytest import MonkeyPatch  # noqa: PT013

from neos_common import error
from neos_common.authorization.base import AccessValidator, SignatureValidator
from neos_common.client import KeycloakClient


class CustomWebClient:
    """Wrapper for httpx.AsyncClient to inject required headers for unit tests."""

    def __init__(self, web_client: httpx.AsyncClient) -> None:
        self._client = web_client

    def _augment_headers(self, kwargs):  # noqa: ANN001, ANN202
        _headers = {"authorization": "bearer token"}
        _headers.update({k.lower(): v for k, v in kwargs.get("headers", {}).items()})
        kwargs["headers"] = _headers
        return kwargs

    async def get(self, *args, **kwargs) -> httpx.Response:
        kwargs = self._augment_headers(kwargs)
        return await self._client.get(*args, **kwargs)

    async def put(self, *args, **kwargs) -> httpx.Response:
        kwargs = self._augment_headers(kwargs)
        return await self._client.put(*args, **kwargs)

    async def post(self, *args, **kwargs) -> httpx.Response:
        kwargs = self._augment_headers(kwargs)
        return await self._client.post(*args, **kwargs)

    async def delete(self, *args, **kwargs) -> httpx.Response:
        kwargs = self._augment_headers(kwargs)
        return await self._client.delete(*args, **kwargs)

    async def request(self, *args, **kwargs) -> httpx.Response:
        kwargs = self._augment_headers(kwargs)
        return await self._client.request(*args, **kwargs)


class BearerWebClient(CustomWebClient):
    ...


class GlobalWebClient(CustomWebClient):
    def _augment_headers(self, kwargs):  # noqa: ANN001, ANN202
        _headers = {"authorization": "bearer token", "x-account": "root", "x-partition": "ksa"}
        _headers.update({k.lower(): v for k, v in kwargs.get("headers", {}).items()})
        kwargs["headers"] = _headers
        return kwargs


@pytest.fixture()
def _auth_patch_factory(  # noqa: PT005
    monkeypatch: MonkeyPatch,
) -> typing.Callable[
    [
        str,
        typing.Union[SignatureValidator, None],
        typing.Union[AccessValidator, None],
        typing.Union[typing.List[str], None],
    ],
    None,
]:
    """Mock token validation checks.

    Mock the keycloak calls required to validate authorization.
    * if provided mock the AccessValidator.validate method to return required resources.
    * if provided mock the SignatureValidator.validate method to return required resources.
    """

    def factory(
        user: str,
        signature_validator: typing.Union[SignatureValidator, None] = None,
        access_validator: typing.Union[AccessValidator, None] = None,
        resources: typing.Union[typing.List[str], None] = None,
    ) -> None:
        monkeypatch.setattr(KeycloakClient, "introspect_token", mock.Mock(return_value={"active": True, "sub": user}))
        # Ignore initial arg (user_id/request) dependent on validator being mocked.
        async def validate(self, _ignore, actions_, resources_, **kwargs):  # noqa: ANN001, ANN202, ARG001
            if resources is None and actions_ != []:
                m = f"The principal <{user}> must have <{actions_[0].value}> action for the resource <{resources_[0].urn}>."  # noqa: E501
                raise error.InsufficientPermissionsError(m)
            return user, resources

        if access_validator:
            monkeypatch.setattr(access_validator, "validate", validate)
        if signature_validator:
            monkeypatch.setattr(signature_validator, "validate", validate)

    return factory



import importlib
import inspect
import re
import typing

from neos_common import schema


def _is_neos_error_code(code: str) -> bool:
    m = re.match(r"^[A-Z]+\d{3}$", code)
    return bool(m)


def get_error_codes(module_names: typing.Union[str, typing.List[str]]) -> schema.ErrorCodes:
    """Get all error codes and messages from the error module."""
    errors = []

    if isinstance(module_names, str):
        module_names = [module_names]

    for module_name in module_names:
        module = importlib.import_module(module_name)

        for name, obj in inspect.getmembers(module):
            if inspect.isclass(obj):
                code = str(getattr(obj, "code", ""))
                message = getattr(obj, "message", "")

                if _is_neos_error_code(code):
                    errors.append(
                        schema.ErrorCode(
                            class_name=name,
                            code=code,
                            message=message,
                        ),
                    )

    return schema.ErrorCodes(errors=errors)



"""Common utils to support neos cli applications."""
import dataclasses
import typing
from csv import DictWriter
from io import StringIO

from fastapi import FastAPI
from yoyo import get_backend, read_migrations
from yoyo.migrations import topological_sort


def migration_upgrade(postgres_dsn: str) -> None:
    backend = get_backend(postgres_dsn)

    migrations = read_migrations("migrations")

    with backend.lock():
        # Apply any outstanding migrations
        backend.apply_migrations(backend.to_apply(migrations))


def migration_downgrade(postgres_dsn: str, count: typing.Union[int, None] = 1) -> None:
    backend = get_backend(postgres_dsn)

    migrations = backend.to_rollback(
        read_migrations("migrations"),
    )
    if count:
        migrations = migrations[:count]

    with backend.lock():
        # Rollback selected migrations (default is latest)
        backend.rollback_migrations(migrations, force=True)


def migration_history(postgres_dsn: str) -> typing.Generator[typing.Tuple[str, str], None, None]:
    backend = get_backend(postgres_dsn)
    migrations = read_migrations("migrations")

    with backend.lock():
        migrations = migrations.__class__(topological_sort(migrations))
        applied = backend.to_rollback(migrations)

    return (("A" if m in applied else " ", m.id) for m in migrations)


@dataclasses.dataclass
class PermissionPair:
    """Permission pair."""

    action: str
    resource: str


@dataclasses.dataclass
class FormattedRoute:
    """Formatted route."""

    methods: str
    path: str
    permission_pairs: typing.List[PermissionPair]


class RoutesPrinter:
    """Routes printer."""

    def __init__(self, app: FastAPI, ignore_routes: typing.List[str]) -> None:
        self.routes = self._get_routes(app, ignore_routes)

    @staticmethod
    def _get_routes(app: FastAPI, ignore_routes: typing.List[str]) -> typing.List[FormattedRoute]:
        routes = []
        for route in app.routes:
            if route.path in ignore_routes:
                continue

            methods = ", ".join(list(route.methods))

            path = route.path

            permission_pairs = []
            if hasattr(route, "openapi_extra") and route.openapi_extra is not None:
                actions = []
                resources = []
                for key in sorted(route.openapi_extra):
                    if key.startswith("x-iam-action"):
                        actions.append(route.openapi_extra[key])
                    if key.startswith("x-iam-resource"):
                        resources.append(route.openapi_extra[key])

                permission_pairs = [PermissionPair(a, r) for a, r in zip(actions, resources)]

            routes.append(FormattedRoute(methods, path, permission_pairs))

        return routes

    def echo(self, echo_fn: typing.Callable) -> None:
        """Print routes to console."""
        for route in self.routes:
            echo_fn(f"{route.methods:<10}{route.path}")
            for permission_pair in route.permission_pairs:
                echo_fn("{:<12}action:   {}".format("", permission_pair.action))
                echo_fn("{:<12}resource: {}".format("", permission_pair.resource))
                echo_fn()
            echo_fn()


class RoutesCSVPrinter(RoutesPrinter):
    """Routes CSV printer."""

    def echo(self, echo_fn: typing.Callable) -> None:
        """Print routes in csv format."""
        csv_output = StringIO()
        writer = DictWriter(csv_output, fieldnames=["methods", "path", "action", "resource"])
        writer.writeheader()

        for route in self.routes:
            header = True
            if len(route.permission_pairs) == 0:
                writer.writerow(
                    {
                        "methods": route.methods if header else "",
                        "path": route.path if header else "",
                        "action": "",
                        "resource": "",
                    },
                )
            else:
                for permission_pair in route.permission_pairs:
                    writer.writerow(
                        {
                            "methods": route.methods if header else "",
                            "path": route.path if header else "",
                            "action": permission_pair.action,
                            "resource": permission_pair.resource,
                        },
                    )
                    header = False

        csv_output.seek(0)
        echo_fn(csv_output.read())
